from __future__ import annotations

__all__ = ["raise_if_missing_deps", "DummyModule"]

from typing import Any

from typing_extensions import Self


def raise_if_missing_deps(*modules: Any) -> None:
    missing_dep_names = []

    for module in modules:
        if isinstance(module, DummyModule):
            missing_dep_names.append(str(module))

    if missing_dep_names:
        raise ImportError(f"Missing modules {missing_dep_names}")


class DummyModule:
    def __init__(self, module_name: str):
        self.__dummy_module__module_name = module_name

    def __str__(self):
        return self.__dummy_module__module_name

    def __repr__(self):
        return f"DummyModule({self.__dummy_module__module_name})"

    def __getattr__(self, item: str) -> Self:
        return self

    def __call__(self, *args, **kwargs) -> Self:
        return self

    def __mro_entries__(self, bases):
        return ()



__all__ = ["maybe_relative_path", "only_one_of"]

from pathlib import Path
from typing import Any, Dict, List, Union

from pydantic import ConfigError, root_validator, validator

from phanas_pydantic_helpers.common.typing import T_MaybeList, ensure_list


def maybe_relative_path(fields: T_MaybeList[str], root_path: Path):
    fields = ensure_list(fields)

    def validate_fn(path: Union[Path, str]):
        if not isinstance(path, Path):
            path = Path(path)
        if not path.is_absolute():
            return root_path / path
        return path

    return validator(*fields, allow_reuse=True)(validate_fn)


def only_one_of(
    *groups_of_fields: T_MaybeList[str], need_all: Union[bool, List[bool]] = True
):
    """
    A Pydantic root validator that ensures one and only one of the groups of
    fields exists in the model.

    :param groups_of_fields: the groups of fields that you want one and only
        one of. Each group may be either a list of field names or a single
        field name.
    :param need_all: whether the groups need all fields or just one to succeed.
        If this arg is a bool, apply to all fields. If it's a list of bools,
        each item will be matched up with the corresponding field group from
        `groups_of_fields` (the list lengths must match).
    """
    if isinstance(need_all, list) and len(need_all) != len(groups_of_fields):
        raise ConfigError(
            "`need_all` must either be a bool or a list of bools of the same "
            "length as `groups_of_fields`"
        )

    groups_of_fields = list(groups_of_fields)
    for idx, group in enumerate(groups_of_fields):
        groups_of_fields[idx] = ensure_list(group)

    def validate_fn(cls, values: Dict[str, Any]):
        a_group_succeeded = False

        for idx, group in enumerate(groups_of_fields):
            required: bool = need_all[idx] if isinstance(need_all, list) else need_all
            existing_fields = [name in values for name in group]

            # Ignore if no fields exist
            if not any(existing_fields):
                continue

            # Two groups coexist, not allowed!
            if a_group_succeeded:
                raise ValueError(
                    f"Only one of the following groups of fields is allowed: "
                    f"{', '.join(map(str, groups_of_fields))}"
                )

            # Check that all fields exist if that was requested
            if required and not all(existing_fields):
                found = [name for name in group if name in values]
                missing = [name for name in group if name not in values]
                raise ValueError(
                    f"The fields {found} exist, but the following are also "
                    f"required and missing: {missing}"
                )

            # If we've made it here, this group has succeeded
            a_group_succeeded = True

        if not a_group_succeeded:
            raise ValueError(
                f"One and only one of the following groups must exist: "
                f"{', '.join(map(str, groups_of_fields))}"
            )

        return values

    return root_validator(pre=True, allow_reuse=True)(validate_fn)



from __future__ import annotations

__all__ = ["create_template_from_model"]

import sys
from typing import TypeVar, get_type_hints

from pydantic import BaseModel

# noinspection PyProtectedMember
from pydantic.main import ModelMetaclass
from typing_extensions import get_args, get_origin

from phanas_pydantic_helpers.common.typing import get_function_args_annotations
from phanas_pydantic_helpers.helpers.field_converter import (
    FieldConverter,
    CONVERTER_METHOD_PREFIX,
)

T = TypeVar("T")

PLACEHOLDER_DICT_KEY_STR = "KEY_NAME"

templatable_types = (list, dict, BaseModel)

MISSING = object()


def create_template_from_type(type_: type[T], field_name: str) -> T:
    if isinstance(type_, ModelMetaclass):
        return create_template_from_model(type_)

    # If a class inherits a `typing.Protocol`, its type will not be `type`
    # like expected, and `issubclass` will not be able to be run on it. We can
    # search for FieldConverter in __mro__ instead.
    if hasattr(type_, "__mro__") and FieldConverter in type_.__mro__:
        converters = (
            (getattr(type_, attr))
            for attr in vars(type_).keys()
            if attr.startswith(CONVERTER_METHOD_PREFIX)
        )
        # We're going to get ONLY the first converter's value type
        try:
            first_converter_type = list(
                get_function_args_annotations(next(converters)).values()
            )[0]
        except StopIteration:
            raise ValueError(
                f'FieldConverter named "{field_name}" has no converter methods'
            )
        return create_template_from_type(
            # Avoid useless type error
            first_converter_type,  # noqa
            field_name,
        )

    if type_ is str:
        return field_name.upper()

    # Get origin type of parameterized type
    origin_type = get_origin(type_)

    if origin_type is dict:
        key_type, value_type = get_args(type_)
        if key_type is str:
            placeholder_key = PLACEHOLDER_DICT_KEY_STR
        else:
            # Get the default value for the key's type
            placeholder_key = key_type()
        return {placeholder_key: create_template_from_type(value_type, field_name)}

    if origin_type is list:
        item_type = get_args(type_)[0]
        return [create_template_from_type(item_type, field_name)]

    # Not parameterized, or some other parameterized type. Return the default
    # value for the type
    return type_()


def create_template_from_model(
    model_type: type[BaseModel] | ModelMetaclass,
) -> dict[str, object]:
    annotations = get_type_hints(
        model_type, vars(sys.modules[model_type.__module__]), vars(model_type)
    )

    dict_ = {}
    for field_name, field in model_type.__fields__.items():
        type_ = annotations.get(field_name, MISSING)

        if not field.required:
            # Field is optional, so it either has a default or a default_factory
            factory = field.default_factory
            if type_ is not MISSING and (
                # Only factories can be templatable types, since they're
                # generated at runtime
                factory in templatable_types
                or isinstance(factory, ModelMetaclass)
            ):
                # This default factory can be templated
                value = create_template_from_type(type_, field_name)
            else:
                # We can't template this default, so just use the default as-is
                value = factory() if factory else field.default

        else:
            # Field is required
            value = create_template_from_type(
                type_ if type is not MISSING else field.type_, field_name
            )

        dict_[field_name] = value

    return dict_



__all__ = ["update_forward_refs_recursive"]

from typing import Type, TypeVar

from pydantic import BaseModel

T_BaseModel = TypeVar("T_BaseModel", bound=Type[BaseModel])


def update_forward_refs_recursive(model: T_BaseModel) -> T_BaseModel:
    for name, value in model.__dict__.items():
        if isinstance(value, type) and issubclass(value, BaseModel):
            update_forward_refs_recursive(value)

    model.update_forward_refs(**model.__dict__)
    return model



__all__ = ["Factory", "instance_list_factory"]

from typing import Callable, List, Type, TypeVar

from pydantic.fields import FieldInfo

V = TypeVar("V")


class Factory(FieldInfo):
    def __init__(self, default_factory, *args, **kwargs):
        super().__init__(*args, default_factory=default_factory, **kwargs)


def instance_list_factory(class_: Type[V], *args, **kwargs) -> Callable[[], List[V]]:
    def make_list():
        return [class_(*args, **kwargs)]

    return make_list



__all__ = [
    "parse_time",
    "parse_duration",
]

import re
from typing import Type, TypeVar

from phanas_pydantic_helpers.common.imports import DummyModule, raise_if_missing_deps

try:
    import pendulum as pen
except ImportError:
    pen = DummyModule("pendulum")

try:
    import pytimeparse
except ImportError:
    pytimeparse = DummyModule("pytimeparse")

time_pattern = re.compile(
    r"^"
    r"(?P<hour>[0-2]?\d)"
    r"(?::?(?P<minute>\d{2}))?"
    r"\s*"
    r"(?:(?P<period_am>a|am)|(?P<period_pm>p|pm))?"
    r"$",
    re.I,
)


V = TypeVar("V")


def parse_time(time_str: str, class_: Type[V] = pen.Time) -> V:
    """
    Parse a string as a time specifier of the general format "12:34 PM".

    :return: the parsed time
    """
    raise_if_missing_deps(pen)

    match = time_pattern.match(time_str)
    if not match:
        raise ValueError("No match")

    hour = int(match["hour"])
    minute = int(match["minute"] or 0)

    if (0 > hour > 23) or (0 > minute > 59):
        raise ValueError("Hour or minute is out of range")

    if match["period_pm"]:
        if hour < 12:
            # This is PM and we use 24 hour times in datetime, so add 12 hours
            hour += 12
        elif hour == 12:
            # 12 PM is 12:00
            pass
        else:
            raise ValueError("24 hour times do not use AM or PM")
    elif match["period_am"]:
        if hour < 12:
            # AM, so no change
            pass
        elif hour == 12:
            # 12 AM is 00:00
            hour = 0
        else:
            raise ValueError("24 hour times do not use AM or PM")

    return class_(hour, minute)


def parse_duration(duration_str: str, class_: Type[V] = pen.Duration) -> V:
    raise_if_missing_deps(pen, pytimeparse)

    seconds = pytimeparse.parse(duration_str)
    if seconds is None:
        raise ValueError("Invalid time duration")
    return class_(seconds=seconds)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['phanas_pydantic_helpers',
 'phanas_pydantic_helpers.common',
 'phanas_pydantic_helpers.fields',
 'phanas_pydantic_helpers.helpers']

package_data = \
{'': ['*']}

install_requires = \
['pydantic>=1.10.4,<2.0.0', 'typing-extensions>=4.4.0,<5.0.0']

extras_require = \
{'time': ['pendulum>=2.1.2,<3.0.0', 'pytimeparse>=1.1.8,<2.0.0']}

setup_kwargs = {
    'name': 'phanas-pydantic-helpers',
    'version': '2.1.3',
    'description': 'A collection of helper functions/classes for Pydantic.',
    'long_description': '# Phana\'s Pydantic Helpers\n\n[![pypi](https://img.shields.io/pypi/v/phanas-pydantic-helpers)]()\n[![pypi-python](https://img.shields.io/pypi/pyversions/phanas-pydantic-helpers)]()\n[![license](https://img.shields.io/github/license/phanabani/phanas-pydantic-helpers)](LICENSE)\n\nA collection of helper functions/classes for Pydantic.\n\n## Table of Contents\n\n- [Install](#install)\n- [Usage](#usage)\n- [Changelog](#changelog)\n- [Developers](#developers)\n- [License](#license)\n\n## Install\n\n### Prerequisites\n\n- [Poetry](https://python-poetry.org/docs/#installation) – dependency manager\n\n### Install Phana\'s Pydantic Helpers\n\nTo get started, install the package with Poetry.\n\n```shell\npoetry add phanas-pydantic-helpers\n```\n\n## Usage\n\n### `Factory`\n\n`Factory(...)` is simply an alias for `pydantic.Field(default_factory=...).`\n\n```python\nfrom pydantic import BaseModel\n\nfrom phanas_pydantic_helpers import Factory\n\n\nclass Config(BaseModel):\n    token: str\n    \n    class _ExtraInfo(BaseModel):\n        name: str = "Unnamed"\n        description: str = "Empty description"\n\n    extra_info: _ExtraInfo = Factory(_ExtraInfo)\n\n\nmodel = Config(token="bleh")\nassert model.extra_info.name == "Unnamed"\nmodel.extra_info.description = "A more detailed description"\n```\n\n### `FieldConverter`\n\nEasily create custom fields with one or more type converters. Make sure the\nfirst superclass is the type you want to represent, as this is considered\nthe main base class and will take precedence over FieldConverter, offering\nbetter code completion.\n\n```python\nfrom phanas_pydantic_helpers import FieldConverter\nfrom pydantic import BaseModel\n\n\nclass ToInt(int, FieldConverter):\n\n    @classmethod\n    def _pyd_convert_str(cls, value: str) -> int:\n        return int(value)\n\n    @classmethod\n    def _pyd_convert_bytes(cls, value: bytes) -> int:\n        return int.from_bytes(value, "big")\n\n\nclass Container(BaseModel):\n    value: ToInt\n\n\ncontainer_from_str = Container(value="5")\nassert container_from_str.value == 5\n\ncontainer_from_bytes = Container(value=b"\\x00\\xFF")\nassert container_from_bytes.value == 0xFF\n```\n\n### `create_template_from_model`\n\nCreate a dict from a model with required fields. This function fills in required\nfields with placeholders.\n\n```python\nfrom typing import Dict, List\n\nfrom pydantic import BaseModel\nfrom phanas_pydantic_helpers import Factory, create_template_from_model\n\n\nclass Player(BaseModel):\n    name: str\n    admin = False\n    highest_score: float = 1.0\n    extra_data: Dict[str, str]\n\n\nclass PlayerDatabase(BaseModel):\n    version: int\n    players: List[Player]\n\n\nclass GameSystem(BaseModel):\n    system_name = "PhanaBox"\n    games: List[str]\n    player_database: PlayerDatabase = Factory(PlayerDatabase)\n\n\nassert create_template_from_model(GameSystem) == {\n    "system_name": "PhanaBox",\n    "games": ["GAMES"],\n    "player_database": {\n        "version": 0,\n        "players": [\n            {\n                "name": "NAME",\n                "admin": False,\n                "highest_score": 1.0,\n                "extra_data": {"KEY_NAME": "EXTRA_DATA"},\n            }\n        ],\n    },\n}\n```\n\n## Changelog\n\nSee [CHANGELOG.md](CHANGELOG.md).\n\n## Developers\n\n### Installation\n\nFollow the installation steps in [install](#install) and use Poetry to \ninstall the development dependencies:\n\n```shell\npoetry install\n```\n\n## License\n\n[MIT © Phanabani.](LICENSE)\n',
    'author': 'Phanabani',
    'author_email': 'phanabani@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/Phanabani/phanas-pydantic-helpers',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import sys
from types import FunctionType, MethodType
from typing import Callable, Dict, List, Optional, TypeVar, Union

from typing_extensions import get_type_hints

__all__ = [
    "T_MaybeList",
    "ensure_list",
    "T_Functionish",
    "get_function_annotations",
    "get_function_args_annotations",
    "get_function_return_type",
]

V = TypeVar("V")
T_MaybeList = Union[V, List[V]]


def ensure_list(value_or_list: T_MaybeList[V]) -> List[V]:
    if not isinstance(value_or_list, list):
        return [value_or_list]
    return value_or_list


T_Functionish = Union[Callable, FunctionType, MethodType, classmethod, staticmethod]


def get_function_annotations(f: T_Functionish):
    try:
        # Unwrap classmethod/staticmethod
        f = f.__func__
    except AttributeError:
        pass
    return get_type_hints(
        f,
        globalns=vars(sys.modules[f.__module__]),
        localns=vars(f),
        include_extras=False,
    )


def get_function_args_annotations(f: T_Functionish) -> Dict[str, type]:
    annotations = get_function_annotations(f)
    try:
        del annotations["return"]
    except KeyError:
        pass
    return annotations


def get_function_return_type(f: T_Functionish) -> Optional[type]:
    annotations = get_function_annotations(f)
    return annotations.get("return")



from __future__ import annotations

__all__ = ["DateTimeField", "TimeField", "DurationField"]

from phanas_pydantic_helpers.common.imports import DummyModule, raise_if_missing_deps
from phanas_pydantic_helpers.common.time import parse_duration, parse_time
from phanas_pydantic_helpers.helpers import FieldConverter

try:
    import pendulum as pen
except ImportError:
    pen = DummyModule("pendulum")


class DateTimeField(pen.DateTime, FieldConverter):
    @classmethod
    def _pyd_convert(cls, datetime_str: str):
        raise_if_missing_deps(pen)

        parsed_dt = pen.parse(datetime_str)
        if not isinstance(parsed_dt, pen.DateTime):
            raise ValueError("Invalid datetime format")
        return parsed_dt


class TimeField(pen.Time, FieldConverter):
    @classmethod
    def _pyd_convert(cls, time_str: str):
        raise_if_missing_deps(pen)
        return parse_time(time_str, class_=cls)


class DurationField(pen.Duration, FieldConverter):
    @classmethod
    def _pyd_convert(cls, duration_str: str):
        raise_if_missing_deps(pen)
        return parse_duration(duration_str, class_=cls)



from . import create_template_from_model as create_template_model_module
from .create_template_from_model import *
from .factory import *
from .field_converter import *
from .misc import *
from .validators import *



__all__ = [
    "CONVERTER_METHOD_PREFIX",
    "FieldConverterError",
    "FieldConverter",
]

from typing import Any, Callable, Dict, Type

from phanas_pydantic_helpers.common.typing import get_function_args_annotations

CONVERTER_METHOD_PREFIX = "_pyd_convert"

T_Converter = Callable[[Type["FieldConverter"], Any], Any]


class FieldConverterError(Exception):
    pass


class FieldConverter:
    """
    Examples:

        class TimezoneField(Timezone, FieldConverter):
            @classmethod
            def _pyd_convert_str(cls, timezone_str: str):
                return cls(timezone_str)

            @classmethod
            def _pyd_convert_timezone(cls, timezone: Timezone):
                return cls(timezone.name)


        class TimeField(pen.Time, FieldConverter):
            @classmethod
            def _pyd_convert(cls, time_str: str):
                return parse_time(time_str, class_=cls)
    """

    __pyd_converters: Dict[type, T_Converter] = None

    @classmethod
    def __get_validators__(cls):
        yield cls.__pyd_convert

    @classmethod
    def __pyd_get_converters(cls) -> Dict[type, T_Converter]:
        if cls.__pyd_converters is not None:
            return cls.__pyd_converters

        converters: Dict[type, T_Converter] = {}
        for name, member in cls.__dict__.items():
            # Iterate through this class's members and find converter methods
            if not isinstance(member, classmethod):
                continue
            if not name.startswith(CONVERTER_METHOD_PREFIX):
                continue

            # Check that the converter method has a single value argument and
            # that the type doesn't already have a converter
            fn = member.__func__  # Unwrap classmethod
            fn_args_types = list(get_function_args_annotations(fn).values())
            if len(fn_args_types) != 1:
                raise FieldConverterError(
                    f"Converter {name} must take one positional argument: value"
                )
            converter_type = fn_args_types[0]
            if converter_type in converters:
                raise FieldConverterError(
                    f"Multiple converters found for type {converter_type}, there "
                    f"should only be one"
                )
            converters[converter_type] = fn

        cls.__pyd_converters = converters
        return cls.__pyd_converters

    @classmethod
    def __pyd_convert(cls, value):
        converters = cls.__pyd_get_converters()
        try:
            fn = converters[type(value)]
        except KeyError:
            raise TypeError(f"No converter for type {type(value)}")
        return fn(cls, value)
